package md3.md3view;

import gl4java.*;
import gl4java.awt.GLAnimCanvas;

import java.awt.event.*;
import java.util.*;

import md3.md3model.*;

/**
 * <p>OpenGL Canvas with functionality to render MD3 models and MD3GLTextures.
 *
 * <p>It will also react to mouseDragged events. If you drag with
 * the first mouse button and don't hold the shift key, you will
 * rotate the model. Dragging the first mouse button while holding down
 * shift will pan the model in the x-y plane. Dragging with the third
 * mouse button scales the model.
 *  
 * @author Erwin Vervaet (erwin@cs.kuleuven.ac.be)
 */
public class MD3GLCanvas extends GLAnimCanvas {
  private static final float MOUSE_ROT_SCALE = 0.5f;
  private static final float MOUSE_PAN_SCALE = 0.1f;
  private static final float MOUSE_ZPOS_SCALE = 0.1f;
  
  private static final double NEAR_GL_PLANE = 0.1d;
  private static final double FAR_GL_PLANE = 512.0d;
    
  private int rendermode, oldrendermode, mode;  
  private int m_x, m_y; //keep track of mouse movements
  private MD3Model model=null; //the top level model that is redered on the canvas  
  private MD3Animation animation=null; //currently played model animation
  private MD3GLTexture texture=null;
  private boolean interpolate;
  
  //cache visitors, to prevent frequent reinstantiation
  private MD3GLModelChangeCurrentFrameVisitor nextVisitor, prevVisitor, rewindVisitor;  	
  
  //display related data
  private float rotAngleX, rotAngleY; //rotation angles
  private float xPos, yPos; //pos in x-y plane, pans model
  private float zPos; //pos on Z axis, scales model
  
  //does the canvas run in OpenGL debug mode?
  private static boolean debug=false;

  //exposed rendering state data 	
  boolean quickManipulations, showBoneFrame, showVertexNormals;
  
  /**
   * <p>Model rendering mode. The current MD3 model will be rendered.
   */
  public static final int MODEL_MODE = 0;
  
  /**
   * <p>Texture rendering mode. The current texture will be rendered.
   */
  public static final int TEXTURE_MODE = 1;
  
  /**
   * <p>Wire frame render mode.
   */
  public static final int WIRE_FRAME = 0;

  /**
   * <p>Flat shaded render mode.
   */
  public static final int FLAT_SHADED = 1;

  /**
   * <p>Flat textured render mode.
   */
  public static final int FLAT_TEXTURED = 2;

  /**
   * <p>Shaded textured render mode.
   */
  public static final int SHADED_TEXTURED = 3;  
  
  //static initializer
  static {
    gl4java.GLContext.gljNativeDebug=debug;
    gl4java.GLContext.gljClassDebug=debug;        
  }

  /**
   * <p>Constructor that creates a new MD3 OpenGL Canvas with the specified
   * width and height. It also sets up the required listeners for the 
   * mouseDragged events.
   *
   * <p>By default, animation and quick manipulations are off, interpolation is on,
   * bone frames or not drawn and the mode is MODEL_MODE.
   *
   * @param width Width of the canvas.
   * @param height Height of the canvas.
   */
  public MD3GLCanvas(int width, int height) {
    super(width,height);
      
    resetManipulations();
    rewindAnimation();
    
    //setup the OpenGL animation canvas
    setVerboseFps(false);    
    setUseRepaint(true);
    setUseFpsSleep(true);    
    
    //default config
    setAnimate(false);
    this.quickManipulations=false;
    this.showBoneFrame=false;
    this.showVertexNormals=false;
    setInterpolate(true);
    setMode(MODEL_MODE);
        
    //register listeners for user manipulations of model view
    
    this.addMouseListener(new MouseAdapter() {
      public void mousePressed(MouseEvent e) {
        m_x=e.getX();
        m_y=e.getY();
        
        if (quickManipulations)
          oldrendermode=rendermode; //save render mode for quickManipulations function
      }
      
      public void mouseReleased(MouseEvent e) {
        if (quickManipulations)
          setRenderMode(oldrendermode); //revert rendermode in case of quickManipulations
            
        if (MD3GLCanvas.this.isSuspended()) //if anim is running, sDisplay will be called by anim thread
        	sDisplay(); 
      }
    });
    
    this.addMouseMotionListener( new MouseMotionAdapter() {
      public void mouseDragged(MouseEvent e) {
        int x=e.getX(),
            y=e.getY();

        if (quickManipulations)
          setRenderMode(WIRE_FRAME);
        
        if (e.getModifiers()==InputEvent.BUTTON1_MASK) { //rotate model
          if (model != null && ((x != m_x) || (y != m_y))) {
            rotAngleX+=((float)(y - m_y)) * MOUSE_ROT_SCALE;
            rotAngleY+=((float)(x - m_x)) * MOUSE_ROT_SCALE;
            if (rotAngleX> 360.0f) rotAngleX-=360.0f;
            if (rotAngleX<-360.0f) rotAngleX+=360.0f;
            if (rotAngleY> 360.0f) rotAngleY-=360.0f;
            if (rotAngleY<-360.0f) rotAngleY+=360.0f;
          }
        }
        else if (e.getModifiers()==(InputEvent.BUTTON1_MASK | InputEvent.SHIFT_MASK)) { //pan model
          if (model != null && ((x != m_x) || (y != m_y))) {
            xPos+=(((float)(x - m_x))/10.0f) * MOUSE_PAN_SCALE;
            yPos+=-1.0f * (((float)(y - m_y))/10.0f) * MOUSE_PAN_SCALE; //switch y orientation
          }
        }
        else if (e.getModifiers()==InputEvent.BUTTON3_MASK) { //scale model
          if (model != null && y != m_y) {
            zPos+=(((float)(y - m_y))/10.0f) * MOUSE_ZPOS_SCALE;
            if (zPos<-1000.0f) zPos=-1000.0f;
            if (zPos>1000.0f) zPos=1000.0f;
          }
        }
        
        if (MD3GLCanvas.this.isSuspended())
        	sDisplay();
        m_x=x;
        m_y=y;
      }
    });
  }
  
  /**
   * <p>Return a GLFunc object that you can use to draw something on this canvas.
   *
   * @return GLFunc object of this GLCanvas.
   */
  protected GLFunc getGL() {
    return this.gl;
  }

  /**
   * <p>Return the canvas's GLUFunc object.
   */  
  protected GLUFunc getGLU() {
  	return this.glu;
  }
  
  /**
   * <p>Override needed when using GLCanvas: render 1 frame, called by paint.
   */
  public void display() {  	
    if( glj.gljMakeCurrent(true) ) {
			this.shallWeRender=true; //make sure we're rendering
      
	    //prepare the canvas for rendering
	    gl.glClear( GLEnum.GL_COLOR_BUFFER_BIT | GLEnum.GL_DEPTH_BUFFER_BIT );
	    gl.glLoadIdentity();
	          
	    //setup canvas according to user manipulations
	    gl.glTranslatef(xPos, yPos, zPos);
	    gl.glScalef( (float)0.05 , (float)0.05, (float)0.05 );
	    gl.glRotatef( rotAngleX, 1.0f ,  0.0f , 0.0f );
	    gl.glRotatef( rotAngleY, 0.0f ,  1.0f , 0.0f );  
      gl.glRotatef( -90.0f , 1.0f ,  0.0f , 0.0f );

      if (mode==MODEL_MODE) {		    							    
	      //draw the model on the canvas
	      if (!isSuspended()) nextFrame();
	      if (model!=null) model.accept(new MD3GLModelDrawModelVisitor(this));
      }
      else if (mode==TEXTURE_MODE) {
      	//draw the texture on the canvas
      	activateRenderMode();
      	MD3GLTextureRenderer.renderTexture(this, texture);
      }

      glj.gljSwap();
      
      if (debug)      
      	glj.gljCheckGL();
    }
  }
  
  /**
   * <p>Override needed when using GLCanvas: resize canvas.
   */
  public void reshape(int width, int height) {
    gl.glViewport(0, 0, width, height);    
    gl.glMatrixMode(GLEnum.GL_PROJECTION); 
    gl.glLoadIdentity();
    if (height > 0)
      glu.gluPerspective((double)90, (double)width/(double)height, NEAR_GL_PLANE, FAR_GL_PLANE);
    gl.glMatrixMode(GLEnum.GL_MODELVIEW);
  }
  
  /** 
   * <p>Override needed when using GLCanvas: create java-stuff and do GL-inits.
   */
  public void init() {
    if( !glj.gljMakeCurrent(true))
      throw new RuntimeException("problem with gljMakeCurrent in init()");
    
    gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    gl.glClearDepth(1.0);
    gl.glDepthFunc(GLEnum.GL_LEQUAL);
    gl.glEnable(GLEnum.GL_DEPTH_TEST);    

    gl.glPixelStorei(GLEnum.GL_UNPACK_ALIGNMENT, 1);  
    gl.glEnable(GLEnum.GL_BLEND);    
    gl.glEnable(GLEnum.GL_NORMALIZE);
       
    float[] mat_specular = { 1.0f, 1.0f, 1.0f, 1.0f };
    float[] mat_shininess = { 20.0f };    
    float[] light_ambient = { 0.9f, 0.9f, 0.9f, 1.0f };
    float[] light0_diffuse = { 1.0f, 1.0f, 1.0f, 1.0f };
    float[] light0_position = { 55.0f, -50.0f, -5.0f, 0.0f };
    float[] light1_diffuse = { 0.5f, 0.5f, 1.0f, 1.0f };
    float[] light1_position = { -50.0f, 45.0f, 15.0f, 0.0f };
  
    gl.glMaterialfv(GLEnum.GL_FRONT, GLEnum.GL_SPECULAR, mat_specular);
    gl.glMaterialfv(GLEnum.GL_FRONT, GLEnum.GL_SHININESS, mat_shininess);
    
    gl.glLightfv(GLEnum.GL_LIGHT0, GLEnum.GL_AMBIENT, light_ambient);
    gl.glLightfv(GLEnum.GL_LIGHT0, GLEnum.GL_DIFFUSE, light0_diffuse);
    gl.glLightfv(GLEnum.GL_LIGHT0, GLEnum.GL_POSITION, light0_position);
    gl.glLightfv(GLEnum.GL_LIGHT1, GLEnum.GL_AMBIENT, light_ambient);
    gl.glLightfv(GLEnum.GL_LIGHT1, GLEnum.GL_DIFFUSE, light1_diffuse);
    gl.glLightfv(GLEnum.GL_LIGHT1, GLEnum.GL_POSITION, light1_position);   
    gl.glEnable(GLEnum.GL_LIGHT0);
    gl.glEnable(GLEnum.GL_LIGHT1);
    gl.glEnable(GLEnum.GL_LIGHTING);

 		gl.glDisable(GLEnum.GL_LINE_SMOOTH);
    gl.glDisable(GLEnum.GL_POLYGON_SMOOTH);    

    gl.glShadeModel(GLEnum.GL_FLAT);   
    setRenderMode(FLAT_TEXTURED);

    reshape(this.getWidth(),this.getHeight()); //setup canvas for perspective view
  }
  
  /** 
   * <p>Override needed when using GLAnimCanvas: re-initialisation after stop for setSuspended(false).
   */
  public void ReInit() {}
  
  /**
   * <p>Reset all user manipulations (scaling + panning + rotating) of this model canvas.
   */
  public void resetManipulations() {
    this.rotAngleX=0;
    this.rotAngleY=0;
    this.xPos=0.0f;
    this.yPos=0.0f;
    this.zPos=-2.0f;
  }

  /**
   * <p>Set the top level model that is rendered on the canvas. This will
   * reset any manipulation that have been done on the canvas (rotate + pan +
   * scale). The canvas will switch to MODEL_MODE.
   *
   * <p>Uploading of texture data to OpenGL is the responsability of the client!
   *
   * @param model The MD3Model to render.
   */  
  public void setModel(MD3Model model) {
    this.model=model;
    resetManipulations();
    setMode(MODEL_MODE);
  }
  
  /**
   * <p>Return the top level model rendered on this canvas.
   *
   * @return The current MD3Model of the canvas.
   */
  public MD3Model getModel() {
    return this.model;
  }
  
  /**
   * <p>Set the texture that will be rendered on the canvas. This will reset any
   * manipulations and switch to TEXTURE_MODE.
   *
   * <p>Uploading of texture data to OpenGL is the responsability of the client!
   */
  public void setTexture(MD3GLTexture texture) {
  	this.texture=texture;
  	resetManipulations();
  	setMode(TEXTURE_MODE);
  }

  /**
   * <p>Return the current texture (rendered in TEXTURE_MODE) of this canvas.
   */  
  public MD3GLTexture getTexture() {
  	return this.texture;
  }
  
  /**
   * <p>Set the operation mode of the canvas. In MODEL_MODE, an MD3Model will be
   * rendered. When in TEXTURE_MODE, a MD3GLTexture object will be shown.
   */
  public void setMode(int mode) {
  	switch (mode) {
  		case TEXTURE_MODE:
  		case MODEL_MODE: 			
  			this.mode=mode;
  			break;
  		default:
	  		throw new IllegalArgumentException(String.valueOf(mode));
  	}
  }
  
  /**
   * <p>Return the current mode of operation of this canvas (TEXTURE_MODE or MODEL_MODE).
   */
  public int getMode() {
  	return this.mode;
  }
  
  /**
   * <p>Start or stop the currently selected animation of the displayed model.
   *
   * @param b true will start the animation, false will stop it.
   */
  public void setAnimate(boolean b) {
  	if (b) {
  		this.threadSuspended=false;	
  		this.setSuspended(false, false); //start thread
  	}
  	else
  		this.threadSuspended=true; //Java 2 way of stopping thread
  }  
  
  /**
   * <p>Set the animation of the model that will be displayed on the canvas.
   *
   * @param anim The animation to play.
   */
  public void setAnimation(MD3Animation anim) {
  	this.animation=anim;
  	
  	//setup visitors to control the animation
  	this.nextVisitor=new MD3GLModelChangeCurrentFrameVisitor(this.animation, MD3GLModelChangeCurrentFrameVisitor.NEXT);
  	this.prevVisitor=new MD3GLModelChangeCurrentFrameVisitor(this.animation, MD3GLModelChangeCurrentFrameVisitor.PREVIOUS);
  	this.rewindVisitor=new MD3GLModelChangeCurrentFrameVisitor(this.animation, MD3GLModelChangeCurrentFrameVisitor.REWIND);
  	
  	rewindAnimation();
    setInterpolate(this.interpolate);
  }
  
  /**
   * <p>Return the currently active animation of the canvas.
   */
  public MD3Animation getAnimation() {
  	return this.animation;
  }
    
  /**
   * <p>Switch to the next frame in the current animation of the model.
   */
  public void nextFrame() {
  	if (model!=null && animation!=null) {
  		nextVisitor.setInterpolate(this.interpolate);
  		model.accept(nextVisitor);
  	}
  }
  
  /**
   * <p>Switch to the previous frame in the current animation.
   */
  public void previousFrame() {
  	if (model!=null && animation!=null) {
  		prevVisitor.setInterpolate(this.interpolate);
  		model.accept(prevVisitor);
  	}
  }
  
  /**
   * <p>Rewind the animation of the model back to the first frame.
   */
  public void rewindAnimation() {
  	if (model!=null && animation!=null)
  		model.accept(rewindVisitor);
  }
  
  /**
   * <p>Enable or disable interpolation between key animation frames.
   */
  public void setInterpolate(boolean b) {
  	this.interpolate=b;
  	//set fps of animation accordingly
  	if (animation!=null) {
	  	if (b)
	  		setAnimateFps((double)(this.animation.fps / MD3GLModelChangeCurrentFrameVisitor.FRACTION));
	  	else
	  		setAnimateFps(animation.fps);
  	}
  }
  
  /**
   * <p>Switch the display to the specified render mode. The supported
   * rendering modes are: WIRE_FRAME, FLAT_SHADED, FLAT_TEXTURED and SHADED_TEXTURED.
   *
   * <p>Note that the new rendermode will only become active after a call to
   * the activateRenderMode() method.
   *
   * <p>This method will have no effect when the canvas is in TEXTURE_MODE.
   *
   * @param rendermode The render mode to swith to.
   */
  public void setRenderMode(int rendermode) {
  	if (mode==MODEL_MODE) {
	  	if (rendermode==WIRE_FRAME || rendermode==FLAT_SHADED || rendermode==FLAT_TEXTURED || rendermode==SHADED_TEXTURED)
	      this.rendermode=rendermode;
	    else
	      throw new IllegalArgumentException(String.valueOf(rendermode));
  	}
  }  
  
  /**
   * <p>Activate the current rendering mode.
   */
  public void activateRenderMode() {
  	//textures allways rendered in textured mode!
  	int rm = (mode==TEXTURE_MODE)?FLAT_TEXTURED:rendermode; 
  	
    switch (rm) {
      case WIRE_FRAME:
        gl.glPolygonMode( GLEnum.GL_FRONT_AND_BACK, GLEnum.GL_LINE );
        gl.glDisable( GLEnum.GL_TEXTURE_2D );
        gl.glDisable( GLEnum.GL_LIGHTING );
        break;
      case FLAT_SHADED:
        gl.glPolygonMode( GLEnum.GL_FRONT_AND_BACK, GLEnum.GL_FILL );
        gl.glDisable( GLEnum.GL_TEXTURE_2D );
        gl.glEnable( GLEnum.GL_LIGHTING );
        break;
      case FLAT_TEXTURED:
        gl.glPolygonMode( GLEnum.GL_FRONT_AND_BACK, GLEnum.GL_FILL );
        gl.glEnable( GLEnum.GL_TEXTURE_2D );
        gl.glDisable( GLEnum.GL_LIGHTING );
        break;
      case SHADED_TEXTURED:       
        gl.glPolygonMode( GLEnum.GL_FRONT_AND_BACK, GLEnum.GL_FILL );
        gl.glEnable( GLEnum.GL_TEXTURE_2D );
        gl.glEnable( GLEnum.GL_LIGHTING );        
        break;
    }
  }
  
  /**
   * <p>Returns the currently used rendering mode.
   */
  public int getRenderMode() {
    return this.rendermode;
  }
  
  /**
   * <p>Dump the image on the canvas to a Targa file.
   *
   * @param filename Name of the output file.
   */
  public void dumpCanvasToTGA(String filename) throws java.io.IOException {
  	byte[] pixels=new byte[this.getWidth() * this.getHeight() * 3];
  	
  	sDisplay(); //make sure the image on the canvas is properly drawn!
  		
  	//get viewport data
		gl.glReadBuffer(GL_BACK);
		gl.glReadPixels(0, 0, this.getWidth(), this.getHeight(), GL_RGB, GL_UNSIGNED_BYTE, pixels);		

		java.io.DataOutputStream fout=new java.io.DataOutputStream(new java.io.FileOutputStream(filename));
		
		//write TGA header
		fout.write(0); //ID length, 0 because no image id field
		fout.write(0); //no color map
		fout.write(2); //image type (24 bit RGB, uncompressed)
		fout.writeShort(0); //color map origin, ignore because no color map
		fout.writeShort(0); //color map length, ignore because no color map
		fout.write(0); //color map entry size, ignore because no color map
		fout.writeShort(0); //x origin
		fout.writeShort(0); //y origin
		fout.write(this.getWidth() & 0x000000FF); fout.write((this.getWidth() & 0x0000FF00) >> 8); //image width (lo-hi)
		fout.write(this.getHeight() & 0x000000FF); fout.write((this.getHeight() & 0x0000FF00) >> 8); //image height (lo-hi)
		fout.write(24); //bpp
		fout.write(0); //description bits
		
		//process image data: TGA pixels should be written in BGR format, so R en B should be switched
		for (int i=0; i<(this.getWidth()*this.getHeight()*3); i+=3) {
			byte tmp=pixels[i];
			pixels[i]=pixels[i+2];
			pixels[i+2]=tmp;
		}
		
		//write TGA image data
		fout.write(pixels, 0, pixels.length);
		
		fout.close();
  }
}