package md3.md3view;

import java.util.*;

import md3.md3model.*;

/**
 * <p>An MD3GLModel visitor that walks through an MD3GLModel structure and applies
 * the specified animation operation (NEXT, PREVIOUS or REWIND) to the animation state
 * data of the encounterd models, taking the specified animation into account.
 *
 * @author Erwin Vervaet (erwin@cs.kuleuven.ac.be) 
 */ 
public class MD3GLModelChangeCurrentFrameVisitor extends MD3GLModelVisitor {

  protected static final float FRACTION = 0.34f;

	/**
	 * <p>Switch to the next frame.
	 */
	public static final int NEXT=0;

	/**
	 * <p>Switch to the previous frame.
	 */
	public static final int PREVIOUS=1;

	/**
	 * <p>Rewind the animation.
	 */
	public static final int REWIND=2;
		
	private MD3Animation anim;
	private int op, upperBound;
	private boolean interpolate;
	
	/**
	 * <p>Create a new visitor to apply an animation operation (NEXT, REWIND, ...)
	 * to a MD3 model. No interpolation will be done.
	 *
	 * @param anim The animation that provides the context for the operation.
	 * @param op The operation to apply.
	 */
	public MD3GLModelChangeCurrentFrameVisitor(MD3Animation anim, int op) {
		this(anim, op, false);
	}
	
	/**
	 * <p>Create a new visitor to apply an animation operation (NEXT, REWIND, ...)
	 * to a MD3 model. The operation is executed in the context of the specified
	 * animation.
	 *
	 * @param anim The animation that provides the context for the operation.
	 * @param op The operation to apply.
	 * @param interpolate Should interpolation be done?
	 */
	public MD3GLModelChangeCurrentFrameVisitor(MD3Animation anim, int op, boolean interpolate) {
		this.anim=anim;
		this.op=op;
		this.interpolate=interpolate;
	}
	
	//calc next frame number
	private int next(int nr) {
		if (nr<(upperBound-1))
			return nr+1;
	  else
	  	return anim.first;
	}
	
	//calc prev frame number
	private int prev(int nr) {
		if (nr==anim.first)
			return upperBound-1;
		else
			return nr-1;
	}
	
	//apply the specified operation to the animation state data members of the model
	//taking the specified animation into account
	private void doOp(MD3GLModel model) {
		//calc upper bound for animation frames in this model
		if (anim.num<0)
			upperBound=model.boneFrameNum;
		else
			upperBound=model.boneFrameNum<(anim.first+anim.num)?model.boneFrameNum:(anim.first+anim.num);
		
		switch (op) {
			case NEXT:
				if (interpolate) {
					model.interpolationFraction+=FRACTION;
					if (model.interpolationFraction>=1.0f) {
						model.currentFrame=model.nextFrame;
						model.nextFrame=next(model.nextFrame);
						model.interpolationFraction=0.0f;
					}
				}
				else {
					model.currentFrame=model.nextFrame;
					model.nextFrame=next(model.nextFrame);
				}
				break;
			case PREVIOUS:
				if (interpolate) {
					model.interpolationFraction-=FRACTION;
					if (model.interpolationFraction<0.0f) {
						model.nextFrame=model.currentFrame;
						model.currentFrame=prev(model.currentFrame);
						model.interpolationFraction=0.8f;
					}
				}
				else {
					model.nextFrame=model.currentFrame;
					model.currentFrame=prev(model.currentFrame);
				}
				break;
			case REWIND:
				model.currentFrame=anim.first;
				model.nextFrame=next(model.currentFrame);
				model.interpolationFraction=0.0f;
				break;
		}
	}
	
	public void visit(MD3GLModel model) {
		if ( (model.loadFilename.toLowerCase().indexOf("lower") != -1 &&
			    (anim.type==AnimationCfg.LEGS || anim.type==AnimationCfg.BOTH))
  		   // this is the LEGS model and the animation is applicable
  		  ||
			   (model.loadFilename.toLowerCase().indexOf("upper") != -1 &&
		      (anim.type==AnimationCfg.TORSO || anim.type==AnimationCfg.BOTH))
			   // this is the TORSO model and the animation is applicable
			  ||
			   anim.type==AnimationCfg.ALL
			   // the animation is allways applicable
			 )
			doOp(model); 
		//else do nothing
		
		//visit children
		Iterator it=model.linkedModels();
		while (it.hasNext())
			((MD3Model)it.next()).accept(this);
	}
	
	/**
	 * <p>Enable or disable interpolation.
	 */
	public void setInterpolate(boolean b) {
		this.interpolate=b;
	}
	
	/**
	 * <p>Is interpolation currently enabled?
	 */
	public boolean getInterpolate() {
		return this.interpolate;
	}
}