package md3.md3view;

import java.io.*;
import java.util.*;

import md3.md3model.*;
import md3.util.*;

/**
 * <p>An MD3GLModel visitor that walks through an MD3GLModel structure and writes
 * the data of the current frame of the encountered models to a specified output stream
 * in Alias Wavefront's Object File (.obj) format (version 3.0).
 *
 * @author Erwin Vervaet (erwin@cs.kuleuven.ac.be) 
 */ 
public class MD3GLModelExportObjVisitor extends MD3GLModelVisitor {
	private StringWriter vertexData;
	private StringWriter elementData;
	private int vertexCount=0,
		          meshCount=0;
  private Stack transformations;
  private float[][] currentTransformation;

  /**
   * <p>Create a new Wavefront object export visitor.
   */
	public MD3GLModelExportObjVisitor() {
		vertexData=new StringWriter();
		elementData=new StringWriter();
    transformations=new Stack();
    //currently no transformation -> 4x4 identity matrix
    currentTransformation=new float[4][4];
    currentTransformation[0][0]=currentTransformation[1][1]=currentTransformation[2][2]=currentTransformation[3][3]=1;
	}
	
  /**
   * <p>Write the generated output data the given output stream.
   *
   * @param outstream The stream to which the data should be written.
   */
	public void writeOut(OutputStream out) throws IOException {
		out.write(vertexData.toString().getBytes());
		out.write(elementData.toString().getBytes());
		out.flush();
	}
	
	private void writeVertexData(MD3Mesh mesh, int frame) {
		for (int i=0; i<mesh.vertexNum; i++) {
      //turn vertex into 4x1 matrix
      float[][] vec=new float[4][1];
      vec[0][0]=mesh.meshFrames[frame][i].x;
      vec[1][0]=mesh.meshFrames[frame][i].y;
      vec[2][0]=mesh.meshFrames[frame][i].z;
      vec[3][0]=1;
      
      //transform vertex
      vec=MatrixMath.mult(currentTransformation, vec);
			
			//write vertex			
			vertexData.write("v " + vec[0][0] + " " + vec[1][0] + " " + vec[2][0] + "\n");
			
			//write texture coord.
			vertexData.write("vt " + mesh.textureCoord[i][0] + " " + mesh.textureCoord[i][1] + "\n");
			
		  //write vertex normal
  		double alpha=mesh.meshVertexNormals[frame][i][0] * 2d * Math.PI / 255d;
  		double beta=mesh.meshVertexNormals[frame][i][1] * 2d * Math.PI / 255d;
		  float a=(float)(Math.cos(beta) * Math.sin(alpha)),
		  	    b=(float)(Math.sin(beta) * Math.sin(alpha)),
		  	    c=(float)Math.cos(alpha);
			
			vertexData.write("vn " + a + " " + b + " " + c + "\n");
			
		}
	}	
		
	private void writeElementData(MD3Mesh mesh) {
		elementData.write("g " + mesh.name + "\n");
		elementData.write("s " + meshCount++ + "\n");
		elementData.write("bevel off\n");
		elementData.write("c_interp off\n");
		elementData.write("d_interp off\n");		
		for (int t=0; t<mesh.triangleNum; t++) {
			int a=vertexCount + mesh.triangles[t][0] + 1,
				  b=vertexCount + mesh.triangles[t][1] + 1,
				  c=vertexCount + mesh.triangles[t][2] + 1;
			elementData.write("f " + a + "/" + a + "/" + a + " " + b + "/" + b + "/" + b + " " + c + "/" + c + "/" + c + "\n");
		}
	}
		
	public void visit(MD3GLModel model) {
		//write meshes
		for (int i=0; i<model.meshNum; i++) {
			MD3Mesh mesh=model.meshes[i];
			writeVertexData(mesh, model.currentFrame);
			writeElementData(mesh);
			
			vertexCount+=mesh.vertexNum;
		}
		
    //write children
    Iterator it=model.linkedModels();
    while (it.hasNext()) {
      MD3Model child=(MD3Model)it.next();
      MD3Tag tag=model.boneFrames[model.currentFrame].tags[child.getParentTagIndex()];
      
      float[][] m=new float[4][4];
      m[0][0] = tag.matrix[0][0]; m[0][1] = tag.matrix[0][1]; m[0][2] = tag.matrix[0][2]; m[0][3] = tag.position.x;
      m[1][0] = tag.matrix[1][0]; m[1][1] = tag.matrix[1][1]; m[1][2] = tag.matrix[1][2]; m[1][3] = tag.position.y;
      m[2][0] = tag.matrix[2][0]; m[2][1] = tag.matrix[2][1]; m[2][2] = tag.matrix[2][2]; m[2][3] = tag.position.z;
      m[3][0] = 0;                m[3][1] = 0;                m[3][2] = 0;                m[3][3] = 1;  
      
      transformations.push(currentTransformation);
      currentTransformation=MatrixMath.mult(currentTransformation,m);
      child.accept(this);
      currentTransformation=(float[][])transformations.pop();
    }
	}
}