package md3.md3view;

import java.io.*;
import java.util.*;

import md3.md3model.*;

/**
 * <p>An MD3Model factory that creates OpenGL compatible MD3 model objects, instead
 * of the default ones. The class itself has a static texture resource manager that will
 * cache read texture data.
 *
 * @author Erwin Vervaet (erwin@cs.kuleuven.ac.be) 
 */ 
public class MD3GLModelFactory extends MD3ModelFactory {
	
  private static transient Map textureRes=new HashMap();
    //map of texture names to MD3GLTexture objects, stores texture resources
    
  public MD3Model makeMD3Model(String name, InputStream in) throws IOException {
  	return new MD3GLModel(name, in);
  }
  
	public MD3Mesh makeMD3Mesh(DataInput din) throws IOException {
		return new MD3GLMesh(din);
	}
	
	/**
   * <p>Factory method that loads a texture and enters it in the texture resource manager.
   *
   * <p>It will use the searchForPath() method of MD3View to resolve the loadFilename, so
   * the texture might be loaded from any data source the MD3View instance has.
   *
   * @param name Name of the texture as specified in .md3 or .skin file.
   * @param loadFilename Full name of file from which texture data will be loaded.
   * @return The loaded MD3GLTexture object or null if there was a problem while loading the texture.
   */
	public MD3Texture makeMD3Texture(String name, String loadFilename) {
		String errorMsg=null;
		loadFilename=loadFilename.trim();
		
		if (loadFilename.length()==0)
    	errorMsg="empty texture filename";
		else
			try {
				if (!(loadFilename.toLowerCase().endsWith(".tga") || loadFilename.toLowerCase().endsWith(".jpg")))
					loadFilename+=".tga";
				
			  String path=MD3View.instance().searchForPath(loadFilename, false);
			  InputStream in=MD3View.instance().getInputStreamForPath(path);
			  MD3Texture res=makeMD3Texture(name, path, in);
			  in.close();
			  return res;
			}
			catch (IOException e) {
				//try other texture file types
				if (MD3ViewOptions.tryAltTexTypes)
					try {
						String ext;
						int i;
						if (loadFilename.toLowerCase().endsWith(".tga")) {
							i=loadFilename.toLowerCase().lastIndexOf(".tga");
							ext=".jpg";
						}
						else {
		          i=loadFilename.toLowerCase().lastIndexOf(".jpg");
		          ext=".tga";
		        }
							
						if (i==-1)
							errorMsg="unable to try " + ext + " version for: " + loadFilename;
						else {
			 				loadFilename=loadFilename.substring(0, i) + ext;
			 				
				  		String path=MD3View.instance().searchForPath(loadFilename, false);
				  		InputStream in=MD3View.instance().getInputStreamForPath(path);
						  MD3Texture res=makeMD3Texture(name, path, in);
						  in.close();
						  return res;
						}
					} 
					catch (IOException ex) {
						errorMsg=ex.getMessage();
					}
				else
					errorMsg=e.getMessage(); 
			}
		
		if (errorMsg!=null && MD3ViewOptions.warningOnTexLoad)
 			MD3View.instance().showExceptionDialog(errorMsg);
			
		return null;
	}
	
	/**
	 * <p>Loads a texture from the given input stream and enters it in the resource manager.
	 */
	public MD3Texture makeMD3Texture(String name, String loadFilename, InputStream in) {
		MD3Texture texture=null;
		
		try {
    	// see if it has been loaded into the texture resource manager
    	texture = (MD3GLTexture)textureRes.get(loadFilename);
  
    	if (texture==null) { // if not we must load a new resource
				texture=new MD3GLTexture(name, loadFilename, in);
	      	
	      // insert texture keyed by file name
	      textureRes.put(loadFilename, texture);
    	}

			//record new name
			texture.name=name;
		} catch (IOException e) {
			if (MD3ViewOptions.warningOnTexLoad)	
	    	MD3View.instance().showExceptionDialog(e.getMessage());
		}
		
		return texture;		
	}
	
	/**
	 * <p>Reloads cached texture data from source for given texture, if present in cache.
	 */
	public static void refreshTexture(String loadFilename) {
		MD3GLTexture cachedTex=(MD3GLTexture)textureRes.get(loadFilename);
			
		if (cachedTex!=null)
			try {
				InputStream in=MD3View.instance().getInputStreamForPath(loadFilename);
				cachedTex.refreshData(in);
				in.close();
			} catch (java.io.IOException e) {
				if (MD3ViewOptions.warningOnTexLoad)	
					MD3View.instance().showExceptionDialog(e.getMessage());
			}
		else
			if (MD3ViewOptions.warningOnTexLoad)	
				MD3View.instance().showExceptionDialog(loadFilename + " not in texture cache");
	}
}