package md3.md3view;

import java.util.*;

import gl4java.*;

import md3.md3model.*;

/**
 * <p>An MD3GLModel visitor that walks through a MD3GLModel structure and
 * uploads all texture data of the visited models into a specified
 * OpenGL context.
 *
 * @author Erwin Vervaet (erwin@cs.kuleuven.ac.be) 
 */
public class MD3GLModelUploadTextureVisitor extends MD3GLModelVisitor {
  private MD3GLCanvas md3canvas;
  
  /**
   * <p>Create a new texture upload visitor that will upload data into
   * the OpenGL context of the specified MD3GLCanvas.
   *
   * @param md3canvas Texture data will be uploaded to this object's OpenGL context.
   */
  public MD3GLModelUploadTextureVisitor(MD3GLCanvas md3canvas) {
    this.md3canvas=md3canvas;    
  }
  
  /**
   * <p>Upload texture information associated with a model into the OpenGL Context.
   *
   * @param model The model of which the textures will be uploaded.
   */
  public void visit(MD3GLModel model) {
    // upload textures of this model to OpenGL
    for (int i=0;i<model.meshNum;i++)
      for (int j=0;j<model.meshes[i].textureNum;j++)
        uploadTextureData(md3canvas.getGL(), (MD3GLTexture)model.meshes[i].textures[j]);
      
    // upload texture data of children
    Iterator it=model.linkedModels();
    while (it.hasNext()) {
      MD3Model child=(MD3Model)it.next();
      child.accept(this);
    }
  }
  
  /**
   * <p>Upload the data of the given texture into the given GL context.
   */
  public static void uploadTextureData(GLFunc gl, MD3GLTexture texture) {
    // set active texture binding and set texture parameters
    gl.glBindTexture( GLEnum.GL_TEXTURE_2D, texture==null?0:texture.bind );
    // filterd texture mode
    gl.glTexParameterf(GLEnum.GL_TEXTURE_2D, GLEnum.GL_TEXTURE_MAG_FILTER, GLEnum.GL_LINEAR);
    gl.glTexParameterf(GLEnum.GL_TEXTURE_2D, GLEnum.GL_TEXTURE_MIN_FILTER, GLEnum.GL_LINEAR);        
    gl.glTexEnvf(GLEnum.GL_TEXTURE_ENV, GLEnum.GL_TEXTURE_ENV_MODE, GLEnum.GL_MODULATE);
    gl.glHint( GLEnum.GL_PERSPECTIVE_CORRECTION_HINT, GLEnum.GL_NICEST );
    // upload the image to OpenGL
    if (texture!=null)
	    gl.glTexImage2D( GLEnum.GL_TEXTURE_2D, 0, 4, texture.width, texture.height, 0, GLEnum.GL_RGBA, GLEnum.GL_UNSIGNED_BYTE, texture.data );

/*
		//mipmapping (only with GL4Java 2.1.2.1 and later!
		gl.glTexParameteri(GLEnum.GL_TEXTURE_2D, GLEnum.GL_TEXTURE_MAG_FILTER, GLEnum.GL_LINEAR);
		gl.glTexParameteri(GLEnum.GL_TEXTURE_2D, GLEnum.GL_TEXTURE_MIN_FILTER, GLEnum.GL_LINEAR_MIPMAP_LINEAR); 
		if (texture!=null)
		  md3canvas.getGLU().gluBuild2DMipmaps(GLEnum.GL_TEXTURE_2D, 4, texture.width, texture.height, GLEnum.GL_RGBA, GLEnum.GL_UNSIGNED_BYTE, texture.data);
*/
  }
}