package md3.util;

/**
 * <p>Booch utility class with quaternion math operations. The quaternions handled
 * by this class are represented as <tt>float[] quat</tt> with <tt>quat.length==4</tt>.
 * The structure of the arrays is as follows: <tt>quat[0]==X</tt>, <tt>quat[1]==Y</tt>,
 * <tt>quat[2]==Z</tt> and <tt>quat[3]==S</tt>, so the scalar part comes last!
 *
 * @author Erwin Vervaet (erwin@cs.kuleuven.ac.be) 
 */ 
public final class QuaternionMath {
  private static final int X=0, Y=1, Z=2, S=3;
  private	static final float DELTA = 0.1f;
  	
	private QuaternionMath() {} //cannot instantiate

  /**
   * <p>Creates a 3x3 rotation matrix from a given quaternion.
   *
   * @param quat The source quaternion.
   * @return The equivalent 3x3 rotation matrix.
   */
  public static final float[][] matrixFromQuaternion(float[] quat) {
  	if (quat.length!=4)
  		throw new IllegalArgumentException("quaternion must be of length 4");
  		
		float[][] res=new float[3][3];		

    // calculate coefficients
    float x2 = quat[X] * quat[X], y2 = quat[Y] * quat[Y], z2 = quat[Z] * quat[Z],
    			xy = quat[X] * quat[Y], xz = quat[X] * quat[Z], yz = quat[Y] * quat[Z],
    			sx = quat[S] * quat[X], sy = quat[S] * quat[Y], sz = quat[S] * quat[Z];
        
    res[0][0] = 1.0f - 2.0f*y2 - 2.0f*z2; res[0][1] = 2.0f*xy - 2.0f*sz;	  res[0][2] = 2.0f*xz + 2.0f*sy;
    res[1][0] = 2.0f*xy + 2.0f*sz; 				res[1][1] = 1.0f - 2.0f*x2 - 2.0f*z2; res[1][2] = 2.0f*yz - 2.0f*sx;
    res[2][0] = 2.0f*xz - 2.0f*sy; 				res[2][1] = 2.0f*yz + 2.0f*sx;	  res[2][2] = 1.0f - 2.0f*x2 - 2.0f*y2;
    
    return res;
  }

  /**
   * <p>Creates a quaternion from a 3x3 rotation matrix.
   *
   * @param mat The 3x3 source rotation matrix.
   * @return The equivalent 4 float quaternion.
   */
  public static final float[] quaternionFromMatrix(float[][] mat) {
  	if (mat.length!=3 || mat[0].length!=3 || mat[1].length!=3 || mat[2].length!=3)
  		throw new IllegalArgumentException("matrix must be 3x3");

  	float[] quat=new float[4];  	

	  // check the diagonal
	  float tr = mat[0][0] + mat[1][1] + mat[2][2] + 1.0f;	
	  if (tr > 0.0625f) {
	    quat[S] = ((float)Math.sqrt((double)tr)) / 2.0f;
	    quat[X] = (mat[1][2] - mat[2][1]) / (4.0f * quat[S]);
	    quat[Y] = (mat[2][0] - mat[0][2]) / (4.0f * quat[S]);
	    quat[Z] = (mat[0][1] - mat[1][0]) / (4.0f * quat[S]);	    
	  } else {
	  	//get biggest diagonal element
	    int i = 0;
	    if (mat[1][1] > mat[0][0]) i = 1;
	    if (mat[2][2] > mat[i][i]) i = 2;
	    //setup index sequence
		  final int[] NXT = {1, 2, 0};  
	    int j = NXT[i];
	    int k = NXT[j];
	
	    float[] q=new float[4];                   
	    q[i] = ((float)Math.sqrt((double)(mat[i][i] - mat[j][j] + mat[k][k] + 1.0f))) / 2.0f;
	    q[j] = (mat[i][j] + mat[j][i]) / (4.0f * q[i]);
	    q[k] = (mat[i][k] + mat[k][i]) / (4.0f * q[i]);
	    q[3] = (mat[j][k] - mat[k][j]) / (4.0f * q[i]);
	
	    quat[X] = q[0];
	    quat[Y] = q[1];
	    quat[Z] = q[2];
	    quat[S] = q[3];
	  }	  
	  
	  return quat;
  }

  /**
   * <p>Do slerp (spherical linear interpolation) interpolation of quaternions along unit 4d sphere.
   *
   * @param from Start rotation.
   * @param to End rotation.   
   * @param frac Specifies the slerp interpolation fraction.
   * @return Quaternion describing rotation interpolated between from and to quaternions.
   */
  public static final float[] quaternionSlerp(float[] from, float[] to, float frac) {
  	if (from.length!=4 || to.length!=4)
  		throw new IllegalArgumentException("quaternion must be of length 4");

		float[] res=new float[4];
	
	  // calc cosine
	  float costheta = from[X]*to[X] + from[Y]*to[Y] + from[Z]*to[Z] + from[S]*to[S];
	
	  // adjust signs (if necessary)
	  boolean flip;
	  if ( costheta < 0.0f ) { 
		  costheta = -costheta;
		  flip=true;
	  } else 
	  	flip=false;
	
	  // calculate coefficients
    float scale0, scale1;
	  if ( (1.0f - costheta) > DELTA ) {
      // standard case (slerp)
      float theta = (float)Math.acos((double)costheta);
      float sintheta = (float)Math.sin((double)theta);
      scale0 = (float)Math.sin((double)((1.0 - frac) * theta)) / sintheta;
      scale1 = (float)Math.sin((double)(frac * theta)) / sintheta;
    } else {        
      // "from" and "to" quaternions are very close 
      //  ... so we can do a linear interpolation
      scale0 = 1.0f - frac;
      scale1 = frac;
    }
   
    if (flip) scale1 = -scale1;
   
	  // calculate final values
	  res[X] = scale0 * from[X] + scale1 * to[X];
	  res[Y] = scale0 * from[Y] + scale1 * to[Y];
	  res[Z] = scale0 * from[Z] + scale1 * to[Z];
	  res[S] = scale0 * from[S] + scale1 * to[S];
	 
	  return res;
  }
}